/*
 *  BINtegrity aka Syswall
 *
 */

#include <linux/fs.h>
#include <linux/elf.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kallsyms.h>
#include <linux/syscalls.h>
#include <linux/uaccess.h>

#include "syswall.h"

#include "hash.h"

#ifdef __LINUX_ARM_ARCH__

struct instruction_t
{
    int op;
    int reg;
    int oper;
    int im;
    int fo;
    int shift;
};

struct state_t
{
    unsigned int regs[16];
};


static inline unsigned int rotl(unsigned int value, int shift) {
    return (value << shift) | (value >> (sizeof(value) * 8 - shift));
}

static inline unsigned int rotr(unsigned int value, int shift) {
    return (value >> shift) | (value << (sizeof(value) * 8 - shift));
}

static inline int arm_disasm_branchlink(struct instruction_t *i, unsigned int ins)
{
    i->op = (ins >> 24) & 0x0f;

    switch (i->op) {
    case 0b1011:
        return 1;
        break;
    }
    return 0;
}

static inline int thumb_disasm_branch(struct instruction_t *i, unsigned short ins)
{
    i->op = (ins >> 11) & 0x0f;
    switch (i->op) {
    case 0b1111:
        printk("thumb long bl\n");
        break;
    }
}

static inline int arm_disasm_data(struct instruction_t *i, unsigned int ins)
{
    i->op = (ins >> 21) & 0x0f;
    i->fo = (ins >> 16) & 0x0f;
    i->reg = (ins >> 12) & 0x0f;
    i->im = (ins >> 25) & 0x1;
    if (i->im) {
        i->oper = ins & 0x0ff;
        i->shift = (ins & 0xf00) >> 8;
        i->oper = rotr(i->oper, i->shift * 2);
    }
    else {
        i->oper = ins & 0x0f;
        i->shift = ins & 0x7f0;
    }
    return 1;
}

static inline int arm_exec(struct state_t *s, struct instruction_t *i)
{
    switch (i->op) {
    case 0b0000:
        if (i->im && s->regs[i->fo] != 0xdeadbeef)
            s->regs[i->reg] = s->regs[i->fo] & i->oper;
        else if (!i->im && s->regs[i->fo] != 0xdeadbeef && s->regs[i->oper] != 0xdeadbeed)
            s->regs[i->reg] = s->regs[i->fo] & s->regs[i->oper];
        break;
    case 0b0001:
        if (i->im && s->regs[i->fo] != 0xdeadbeef)
            s->regs[i->reg] = s->regs[i->fo] ^ i->oper;
        else if (!i->im && s->regs[i->fo] != 0xdeadbeef && s->regs[i->oper] != 0xdeadbeed)
            s->regs[i->reg] = s->regs[i->fo] ^ s->regs[i->oper];
        break;
    case 0b1100:
        if (i->im && s->regs[i->fo] != 0xdeadbeef)
            s->regs[i->reg] = s->regs[i->fo] | i->oper;
        else if (!i->im && s->regs[i->fo] != 0xdeadbeef && s->regs[i->oper] != 0xdeadbeed)
            s->regs[i->reg] = s->regs[i->fo] | s->regs[i->oper];
        break;
    case 0b0010:
        if (i->im && s->regs[i->fo] != 0xdeadbeef)
            s->regs[i->reg] = s->regs[i->fo] - i->oper;
        else if (!i->im && s->regs[i->fo] != 0xdeadbeef && s->regs[i->oper] != 0xdeadbeed)
            s->regs[i->reg] = s->regs[i->fo] - s->regs[i->oper];
        break;
    case 0b0100:
        if (i->im && s->regs[i->fo] != 0xdeadbeef)
            s->regs[i->reg] = s->regs[i->fo] + i->oper;
        else if (!i->im && s->regs[i->fo] != 0xdeadbeef && s->regs[i->oper] != 0xdeadbeed)
            s->regs[i->reg] = s->regs[i->fo] + s->regs[i->oper];
        break;
    case 0b1101:
        if (i->im)
            s->regs[i->reg] = i->oper;
        else if (!i->im && s->regs[i->oper] != 0xdeadbeef)
            s->regs[i->reg] = s->regs[i->oper];
        break;
    default:
        s->regs[i->reg] = 0xdeadbeef;
        break;
    }   
}

#define SNUMINS 12
#define SNUMINSB (SNUMINS*4)

int vet_subreturn(struct file *f, unsigned offset, unsigned int *argv, unsigned int argc, unsigned int regmodlist, struct vetmap_t **vmp)
{
    char buffer[SNUMINSB];
    char *buf = &buffer[SNUMINSB - 8];
	int thumb = offset % 4;
    int retval = -2;
    struct state_t s;
    int i;

    for (i = 0; i < 16; i++)
        s.regs[i] = 0xdeadbeef;

	if (thumb) {
		offset -= 3;
	    kernel_read(f, offset-((SNUMINS-1)*4), buffer, SNUMINSB);
		
		if (buf[3] == 0xF7 || buf[3] == 0xF6 || buf[3] == 0xF0 || buf[3] == 0xF5) {
			retval = 0;
		}
		else {
			printk("%x %x %x %x\n", buf[0], buf[1], buf[2], buf[3]);
			printk("%x %x %x %x\n", buf[4], buf[5], buf[6], buf[7]);
            return retval;
		}
	}
	else {
	    kernel_read(f, offset-((SNUMINS-1)*4), buffer, SNUMINSB);
	
		if (buf[3] == 0xFA || buf[3] == 0xEB || buf[3] == 0xBA || buf[3] == 0x9A || buf[3] == 0xDA || buf[3] == 0x0B || buf[3] == 0xE1) {
			retval = 0;
    
            struct instruction_t is;
            unsigned int *ins = (unsigned int*) buffer;
            for (i = 0; i < SNUMINS - 2; i++) {
                arm_disasm_data(&is, ins[i]);
                arm_exec(&s, &is);
            }
        }
		else {
			printk("%x %x %x %x\n", buf[0], buf[1], buf[2], buf[3]);
			printk("%x %x %x %x\n", buf[4], buf[5], buf[6], buf[7]);
            return retval;
		}
	}

    struct vetmap_t *vm = *vmp;
    for (unsigned i = 0; i < argc; i++) {
        if (regmodlist & (1<<i) || s.regs[i] == 0xdeadbeef) {
            vm->argv[i] = 0xdeadbeef;
            continue;
        }
        if (s.regs[i] != 0xdeadbeef && argv[i] == s.regs[i]) {
            vm->argv[i] = s.regs[i];
        }
        else {
            retval = -1;
            break;
        }
    }

	return retval;
}

static inline int arm_disasm_outreg(void *i, unsigned int ins)
{
    int reg;
    int op;

    if ((ins & 0x0ffffff0) == 0x012fff10)
        return -1;
    if ((ins & 0x0f000000) == 0x0f000000)
        return -1;

    reg = (ins >> 12) & 0x0f;
    return reg;
}

#define NUMINS 20
#define NUMINSB (NUMINS*4)

int vet_syscall(struct file *f, unsigned offset, unsigned syscall_number, unsigned int *regmodlist)
{
    DPRINT(KERN_INFO "\tvet_syscall\n");

    int retval = -2;
    char buffer[NUMINSB];
    char *buf = &buffer[NUMINSB - 28];

    kernel_read(f, offset-((NUMINS-5)*4), buffer, NUMINSB);
   
    if (buf[7] == 0xef && buf[6] == 0 && buf[5] == 0 && buf[4] == 0) {
        if (buf[3] == 0xe3) {
            struct instruction_t is;
            struct state_t s;

            s.regs[7] = 0xdeadbeef;

            unsigned int *ins = (unsigned int*) buf;
            arm_disasm_data(&is, ins[0]);
            arm_exec(&s, &is);

            if (s.regs[7] == syscall_number)
                retval = 0;
            else
                printk("vet_syscall r7 = %#x\n", s.regs[7]);
        }
        else if (buf[3] == 0xe5 && buf[2] == 0x9f && buf[1] == 0x70) {
            if (buf[0] == 16) {
                int fsc = (buf[27] << 24) | (buf[26] << 16) | (buf[25] << 8) | buf[24];
                if (fsc == syscall_number) {
                    retval = 0;
                }
            }
            else {
                printk(KERN_INFO "LDR (offset != 16): %x %x %x %x\n", buf[24], buf[25], buf[26], buf[27]);
            }        
        }
        else {
            printk(KERN_INFO "unexpected instruction: %x %x %x %x at %x\n", buf[0], buf[1], buf[2], buf[3], offset-8);
            if (buf[3] == 0xe8 && buf[2] == 0x9c && buf[1] == 0x00 && buf[0] == 0x78) {
                printk(KERN_INFO "likely this is the generic syscall in libc, we can't verify the scno\n");

                printk("Syscall by %s pid %d, tgid %d\n", current->comm, current->pid, current->tgid);
                printk("Syscall #%u\n", syscall_number);
                printk("file %s offset : %#x\n", f->f_path.dentry->d_name.name, offset);

                if (syscall_number != 192 || syscall_number != 292)
                    retval = -1;

                retval = 0;
            }
        }
    }
    else {
        printk("%x %x %x %x\n", buf[0], buf[1], buf[2], buf[3]);
        printk("%x %x %x %x\n", buf[4], buf[5], buf[6], buf[7]);
        
        printk("file %s offset : %#x\n", f->f_path.dentry->d_name.name, offset);
        retval = -2;
    }

    if (retval == 0) {
        unsigned int *bc = (unsigned int*) buffer;
        for (int i = 0; i < (NUMINS - 7); i++) {
            int r = arm_disasm_outreg(0, bc[i]);
            switch (r) {
            case 14:
                *regmodlist = *regmodlist | MOD_REG_RA;
                retval = -3;
                break;
            case 0:
                *regmodlist = *regmodlist | MOD_REG_A1;
                break;
            case 1:
                *regmodlist = *regmodlist | MOD_REG_A2;
                break;
            case 2:
                *regmodlist = *regmodlist | MOD_REG_A3;
                break;
            case 3:
                *regmodlist = *regmodlist | MOD_REG_A4;
                break;
            }
        }
        if ((*regmodlist) & 0x7)
            *regmodlist = *regmodlist | MOD_REG_A4;
    }

    return retval;
}

#define __flush_icache_all_generic()                                    \
        asm("mcr        p15, 0, %0, c7, c5, 0"                          \
             : : "r" (0));

/* Invalidate I-cache inner shareable */
#define __flush_icache_all_v7_smp()                                     \
        asm("mcr        p15, 0, %0, c7, c1, 0"                          \
             : : "r" (0));



#define flush_icache() { \
        __asm__("mov r0, #0\n"\
    "mcr p15, 0, r0, c7, c5, 0\n"\
    "mov r0, r0\n"\
    "mcrne  p15, 0, ip, c7, c5, 0\n"\
    ); \
}

void disable_page_protection(void) {}

void enable_page_protection(void) 
{
    flush_icache();
    __flush_icache_all_generic();
    __flush_icache_all_v7_smp();
}

void hook_arm_sys_wrapper(char *name, char *rname, unsigned char *check)
{
    unsigned char *nameaddr = (unsigned char*) kallsyms_lookup_name(name);
    unsigned char *sfw = nameaddr;
    
    if (!nameaddr)
        return;

    memcpy(check + (9*4), nameaddr, 6*4);
    if (*(check+(10*4)+3) == 0xea) {
        memcpy(check + (10*4), check+(16*4), 4);
        sfw = (unsigned char*) kallsyms_lookup_name(rname);
        memcpy(check + (11*4), &sfw, 4);
    }
    else {
        memcpy(check + (11*4), check+(16*4), 4);   
        sfw = (unsigned char*) kallsyms_lookup_name(rname);
        memcpy(check + (12*4), &sfw, 4);
    }

    memcpy(nameaddr, check + (16*4), 2*4);
}

#endif
